// Copyright 2020 ConsenSys AG
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Package plonk implements PLONK Zero Knowledge Proof system.
//
// See also
//
// https://eprint.iacr.org/2019/953
package plonk

import (
	"io"

	"github.com/consensys/gnark-crypto/ecc"
	"github.com/consensys/gnark-crypto/kzg"
	"github.com/consensys/gnark/backend"
	"github.com/consensys/gnark/frontend"

	cs_bls12377 "github.com/consensys/gnark/internal/backend/bls12-377/cs"
	cs_bls12381 "github.com/consensys/gnark/internal/backend/bls12-381/cs"
	cs_bls24315 "github.com/consensys/gnark/internal/backend/bls24-315/cs"
	cs_bn254 "github.com/consensys/gnark/internal/backend/bn254/cs"
	cs_bw6761 "github.com/consensys/gnark/internal/backend/bw6-761/cs"

	plonk_bls12377 "github.com/consensys/gnark/internal/backend/bls12-377/plonk"
	plonk_bls12381 "github.com/consensys/gnark/internal/backend/bls12-381/plonk"
	plonk_bls24315 "github.com/consensys/gnark/internal/backend/bls24-315/plonk"
	plonk_bn254 "github.com/consensys/gnark/internal/backend/bn254/plonk"
	plonk_bw6761 "github.com/consensys/gnark/internal/backend/bw6-761/plonk"

	witness_bls12377 "github.com/consensys/gnark/internal/backend/bls12-377/witness"
	witness_bls12381 "github.com/consensys/gnark/internal/backend/bls12-381/witness"
	witness_bls24315 "github.com/consensys/gnark/internal/backend/bls24-315/witness"
	witness_bn254 "github.com/consensys/gnark/internal/backend/bn254/witness"
	witness_bw6761 "github.com/consensys/gnark/internal/backend/bw6-761/witness"

	kzg_bls12377 "github.com/consensys/gnark-crypto/ecc/bls12-377/fr/kzg"
	kzg_bls12381 "github.com/consensys/gnark-crypto/ecc/bls12-381/fr/kzg"
	kzg_bls24315 "github.com/consensys/gnark-crypto/ecc/bls24-315/fr/kzg"
	kzg_bn254 "github.com/consensys/gnark-crypto/ecc/bn254/fr/kzg"
	kzg_bw6761 "github.com/consensys/gnark-crypto/ecc/bw6-761/fr/kzg"
)

// Proof represents a Plonk proof generated by plonk.Prove
//
// it's underlying implementation is curve specific (see gnark/internal/backend)
type Proof interface {
	io.WriterTo
	io.ReaderFrom
}

// ProvingKey represents a plonk ProvingKey
//
// it's underlying implementation is strongly typed with the curve (see gnark/internal/backend)
type ProvingKey interface {
	io.WriterTo
	io.ReaderFrom
	InitKZG(srs kzg.SRS) error
	VerifyingKey() interface{}
}

// VerifyingKey represents a plonk VerifyingKey
//
// it's underlying implementation is strongly typed with the curve (see gnark/internal/backend)
type VerifyingKey interface {
	io.WriterTo
	io.ReaderFrom
	InitKZG(srs kzg.SRS) error
	NbPublicWitness() int // number of elements expected in the public witness
}

// Setup prepares the public data associated to a circuit + public inputs.
func Setup(ccs frontend.CompiledConstraintSystem, kzgSRS kzg.SRS) (ProvingKey, VerifyingKey, error) {

	switch tccs := ccs.(type) {
	case *cs_bn254.SparseR1CS:
		return plonk_bn254.Setup(tccs, kzgSRS.(*kzg_bn254.SRS))
	case *cs_bls12381.SparseR1CS:
		return plonk_bls12381.Setup(tccs, kzgSRS.(*kzg_bls12381.SRS))
	case *cs_bls12377.SparseR1CS:
		return plonk_bls12377.Setup(tccs, kzgSRS.(*kzg_bls12377.SRS))
	case *cs_bw6761.SparseR1CS:
		return plonk_bw6761.Setup(tccs, kzgSRS.(*kzg_bw6761.SRS))
	case *cs_bls24315.SparseR1CS:
		return plonk_bls24315.Setup(tccs, kzgSRS.(*kzg_bls24315.SRS))
	default:
		panic("unrecognized SparseR1CS curve type")
	}

}

// Prove generates PLONK proof from a circuit, associated preprocessed public data, and the witness
// if the force flag is set:
// 	will executes all the prover computations, even if the witness is invalid
//  will produce an invalid proof
//	internally, the solution vector to the SparseR1CS will be filled with random values which may impact benchmarking
func Prove(ccs frontend.CompiledConstraintSystem, pk ProvingKey, fullWitness frontend.Circuit, opts ...func(opt *backend.ProverOption) error) (Proof, error) {

	// apply options
	opt, err := backend.NewProverOption(opts...)
	if err != nil {
		return nil, err
	}

	switch tccs := ccs.(type) {
	case *cs_bn254.SparseR1CS:
		w := witness_bn254.Witness{}
		if err := w.FromFullAssignment(fullWitness); err != nil {
			return nil, err
		}
		return plonk_bn254.Prove(tccs, pk.(*plonk_bn254.ProvingKey), w, opt)

	case *cs_bls12381.SparseR1CS:
		w := witness_bls12381.Witness{}
		if err := w.FromFullAssignment(fullWitness); err != nil {
			return nil, err
		}
		return plonk_bls12381.Prove(tccs, pk.(*plonk_bls12381.ProvingKey), w, opt)

	case *cs_bls12377.SparseR1CS:
		w := witness_bls12377.Witness{}
		if err := w.FromFullAssignment(fullWitness); err != nil {
			return nil, err
		}
		return plonk_bls12377.Prove(tccs, pk.(*plonk_bls12377.ProvingKey), w, opt)

	case *cs_bw6761.SparseR1CS:
		w := witness_bw6761.Witness{}
		if err := w.FromFullAssignment(fullWitness); err != nil {
			return nil, err
		}
		return plonk_bw6761.Prove(tccs, pk.(*plonk_bw6761.ProvingKey), w, opt)

	case *cs_bls24315.SparseR1CS:
		w := witness_bls24315.Witness{}
		if err := w.FromFullAssignment(fullWitness); err != nil {
			return nil, err
		}
		return plonk_bls24315.Prove(tccs, pk.(*plonk_bls24315.ProvingKey), w, opt)

	default:
		panic("unrecognized SparseR1CS curve type")
	}
}

// Verify verifies a PLONK proof, from the proof, preprocessed public data, and public witness.
func Verify(proof Proof, vk VerifyingKey, publicWitness frontend.Circuit) error {

	switch _proof := proof.(type) {

	case *plonk_bn254.Proof:
		w := witness_bn254.Witness{}
		if err := w.FromPublicAssignment(publicWitness); err != nil {
			return err
		}
		return plonk_bn254.Verify(_proof, vk.(*plonk_bn254.VerifyingKey), w)

	case *plonk_bls12381.Proof:
		w := witness_bls12381.Witness{}
		if err := w.FromPublicAssignment(publicWitness); err != nil {
			return err
		}
		return plonk_bls12381.Verify(_proof, vk.(*plonk_bls12381.VerifyingKey), w)

	case *plonk_bls12377.Proof:
		w := witness_bls12377.Witness{}
		if err := w.FromPublicAssignment(publicWitness); err != nil {
			return err
		}
		return plonk_bls12377.Verify(_proof, vk.(*plonk_bls12377.VerifyingKey), w)

	case *plonk_bw6761.Proof:
		w := witness_bw6761.Witness{}
		if err := w.FromPublicAssignment(publicWitness); err != nil {
			return err
		}
		return plonk_bw6761.Verify(_proof, vk.(*plonk_bw6761.VerifyingKey), w)

	case *plonk_bls24315.Proof:
		w := witness_bls24315.Witness{}
		if err := w.FromPublicAssignment(publicWitness); err != nil {
			return err
		}
		return plonk_bls24315.Verify(_proof, vk.(*plonk_bls24315.VerifyingKey), w)

	default:
		panic("unrecognized proof type")
	}
}

// NewCS instantiate a concrete curved-typed SparseR1CS and return a CompiledConstraintSystem interface
// This method exists for (de)serialization purposes
func NewCS(curveID ecc.ID) frontend.CompiledConstraintSystem {
	var r1cs frontend.CompiledConstraintSystem
	switch curveID {
	case ecc.BN254:
		r1cs = &cs_bn254.SparseR1CS{}
	case ecc.BLS12_377:
		r1cs = &cs_bls12377.SparseR1CS{}
	case ecc.BLS12_381:
		r1cs = &cs_bls12381.SparseR1CS{}
	case ecc.BW6_761:
		r1cs = &cs_bw6761.SparseR1CS{}
	case ecc.BLS24_315:
		r1cs = &cs_bls24315.SparseR1CS{}
	default:
		panic("not implemented")
	}
	return r1cs
}

// NewProvingKey instantiates a curve-typed ProvingKey and returns an interface
// This function exists for serialization purposes
func NewProvingKey(curveID ecc.ID) ProvingKey {
	var pk ProvingKey
	switch curveID {
	case ecc.BN254:
		pk = &plonk_bn254.ProvingKey{}
	case ecc.BLS12_377:
		pk = &plonk_bls12377.ProvingKey{}
	case ecc.BLS12_381:
		pk = &plonk_bls12381.ProvingKey{}
	case ecc.BW6_761:
		pk = &plonk_bw6761.ProvingKey{}
	case ecc.BLS24_315:
		pk = &plonk_bls24315.ProvingKey{}
	default:
		panic("not implemented")
	}

	return pk
}

// NewProof instantiates a curve-typed ProvingKey and returns an interface
// This function exists for serialization purposes
func NewProof(curveID ecc.ID) Proof {
	var proof Proof
	switch curveID {
	case ecc.BN254:
		proof = &plonk_bn254.Proof{}
	case ecc.BLS12_377:
		proof = &plonk_bls12377.Proof{}
	case ecc.BLS12_381:
		proof = &plonk_bls12381.Proof{}
	case ecc.BW6_761:
		proof = &plonk_bw6761.Proof{}
	case ecc.BLS24_315:
		proof = &plonk_bls24315.Proof{}
	default:
		panic("not implemented")
	}

	return proof
}

// NewVerifyingKey instantiates a curve-typed VerifyingKey and returns an interface
// This function exists for serialization purposes
func NewVerifyingKey(curveID ecc.ID) VerifyingKey {
	var vk VerifyingKey
	switch curveID {
	case ecc.BN254:
		vk = &plonk_bn254.VerifyingKey{}
	case ecc.BLS12_377:
		vk = &plonk_bls12377.VerifyingKey{}
	case ecc.BLS12_381:
		vk = &plonk_bls12381.VerifyingKey{}
	case ecc.BW6_761:
		vk = &plonk_bw6761.VerifyingKey{}
	case ecc.BLS24_315:
		vk = &plonk_bls24315.VerifyingKey{}
	default:
		panic("not implemented")
	}

	return vk
}

// ReadAndProve generates PLONK proof from a circuit, associated proving key, and the full witness
func ReadAndProve(ccs frontend.CompiledConstraintSystem, pk ProvingKey, witness io.Reader, opts ...func(opt *backend.ProverOption) error) (Proof, error) {

	// apply options
	opt, err := backend.NewProverOption(opts...)
	if err != nil {
		return nil, err
	}

	_, nbSecret, nbPublic := ccs.GetNbVariables()
	expectedSize := (nbSecret + nbPublic)

	switch tccs := ccs.(type) {
	case *cs_bn254.SparseR1CS:
		_pk := pk.(*plonk_bn254.ProvingKey)
		w := witness_bn254.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return nil, err
		}
		proof, err := plonk_bn254.Prove(tccs, _pk, w, opt)
		if err != nil {
			return proof, err
		}
		return proof, nil

	case *cs_bls12381.SparseR1CS:
		_pk := pk.(*plonk_bls12381.ProvingKey)
		w := witness_bls12381.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return nil, err
		}
		proof, err := plonk_bls12381.Prove(tccs, _pk, w, opt)
		if err != nil {
			return proof, err
		}
		return proof, nil

	case *cs_bls12377.SparseR1CS:
		_pk := pk.(*plonk_bls12377.ProvingKey)
		w := witness_bls12377.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return nil, err
		}
		proof, err := plonk_bls12377.Prove(tccs, _pk, w, opt)
		if err != nil {
			return proof, err
		}
		return proof, nil

	case *cs_bw6761.SparseR1CS:
		_pk := pk.(*plonk_bw6761.ProvingKey)
		w := witness_bw6761.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return nil, err
		}
		proof, err := plonk_bw6761.Prove(tccs, _pk, w, opt)
		if err != nil {
			return proof, err
		}
		return proof, nil

	case *cs_bls24315.SparseR1CS:
		_pk := pk.(*plonk_bls24315.ProvingKey)
		w := witness_bls24315.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return nil, err
		}
		proof, err := plonk_bls24315.Prove(tccs, _pk, w, opt)
		if err != nil {
			return proof, err
		}
		return proof, nil

	default:
		panic("unrecognized R1CS curve type")
	}
}

// ReadAndVerify verifies a PLONK proof from a circuit, associated proving key, and the full witness
func ReadAndVerify(proof Proof, vk VerifyingKey, witness io.Reader) error {

	expectedSize := vk.NbPublicWitness()

	switch _proof := proof.(type) {
	case *plonk_bn254.Proof:
		_vk := vk.(*plonk_bn254.VerifyingKey)
		w := witness_bn254.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return err
		}
		return plonk_bn254.Verify(_proof, _vk, w)

	case *plonk_bls12381.Proof:
		_vk := vk.(*plonk_bls12381.VerifyingKey)
		w := witness_bls12381.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return err
		}
		return plonk_bls12381.Verify(_proof, _vk, w)

	case *plonk_bls12377.Proof:
		_vk := vk.(*plonk_bls12377.VerifyingKey)
		w := witness_bls12377.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return err
		}
		return plonk_bls12377.Verify(_proof, _vk, w)

	case *plonk_bw6761.Proof:
		_vk := vk.(*plonk_bw6761.VerifyingKey)
		w := witness_bw6761.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return err
		}
		return plonk_bw6761.Verify(_proof, _vk, w)

	case *plonk_bls24315.Proof:
		_vk := vk.(*plonk_bls24315.VerifyingKey)
		w := witness_bls24315.Witness{}
		if _, err := w.LimitReadFrom(witness, expectedSize); err != nil {
			return err
		}
		return plonk_bls24315.Verify(_proof, _vk, w)

	default:
		panic("unrecognized R1CS curve type")
	}
}

// IsSolved attempts to solve the constraint system with provided witness
// returns nil if it succeeds, error otherwise.
func IsSolved(ccs frontend.CompiledConstraintSystem, witness frontend.Circuit, opts ...func(opt *backend.ProverOption) error) error {

	opt, err := backend.NewProverOption(opts...)
	if err != nil {
		return err
	}

	switch tccs := ccs.(type) {
	case *cs_bn254.SparseR1CS:
		w := witness_bn254.Witness{}
		if err := w.FromFullAssignment(witness); err != nil {
			return err
		}
		return tccs.IsSolved(w, opt)
	case *cs_bls12381.SparseR1CS:
		w := witness_bls12381.Witness{}
		if err := w.FromFullAssignment(witness); err != nil {
			return err
		}
		return tccs.IsSolved(w, opt)
	case *cs_bls12377.SparseR1CS:
		w := witness_bls12377.Witness{}
		if err := w.FromFullAssignment(witness); err != nil {
			return err
		}
		return tccs.IsSolved(w, opt)
	case *cs_bw6761.SparseR1CS:
		w := witness_bw6761.Witness{}
		if err := w.FromFullAssignment(witness); err != nil {
			return err
		}
		return tccs.IsSolved(w, opt)
	case *cs_bls24315.SparseR1CS:
		w := witness_bls24315.Witness{}
		if err := w.FromFullAssignment(witness); err != nil {
			return err
		}
		return tccs.IsSolved(w, opt)
	default:
		panic("unknown constraint system type")
	}
}
